/*
  
 Graph Class Implementation File
 Created by: Alejandro Gutierrez
 Created Date: 08/05/2007
 Last Modified Date: 07/05/2008
 University of South Florida

 Copyright (C) 2008  Alejandro Gutierrez Munoz

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "graph.h"

#include <sstream>
#include <queue>
#include <cmath>

#include <jama_lu.h> 
#include <tnt.h>

graph::graph(void){
}

graph::graph(const string &s,map< string,double > &label_counts, double &max_resistance, double &min_resistance)
{
	try {

		int n,m;
		char c;
		string str;
		pair< int,float > list_node;
		pair< int,string > graph_vertice;
		istringstream graph_string(s);
		istringstream buffer;
		string line;

		if (!s.empty()) {

			// get first line and check if it is BEGIN
			getline(graph_string,line);
			buffer.str(line);

			if ( line.compare(0,5,"BEGIN") ) {
				throw (int) __LINE__;
			}

			buffer >> str; // skip the BEGIN
			buffer >> this->graph_name;  // name
			buffer >> this->graph_class;  // class

			buffer.clear();

			getline(graph_string,line);		

			while (!graph_string.eof() && line.compare(0,3,"END")  ) {
													
				buffer.str(line);

				if (line.at(0) == 'v' || line.at(0) == 'V') {

					// read vertice line
					buffer >> c;  //char 'v'
					buffer >> n;   // vertex id
					buffer >> str;  // label

					graph_vertice.first = n-1;
					graph_vertice.second = str;

					// add vertex to the list of vertices at the end of the array
					this->vertices.push_back(graph_vertice); 

					// add label to label counts
					if (!label_counts[graph_vertice.second]) {
						label_counts[graph_vertice.second] = 1;
					}
					else {
						label_counts[graph_vertice.second] = label_counts[graph_vertice.second] + 1;
					}

				}
				else if (line.at(0) == 'e' || line.at(0) == 'E') {

					// read edge line
					buffer >> c; // char 'e'
					buffer >> n; // "from"  vertex
					buffer >> m; // "to"  vertex
					buffer >> str;  // weight

																		
					//check if the size of the list is 0. If so, this means that it's the first
					//edge checked. Proceed to create the list
			
					if ( this->adjacency_list.size() == 0) {
						// create a vector with size |v|.
						this->adjacency_list.resize(this->vertices.size());
					}

					if (n > adjacency_list.size() || m > adjacency_list.size() ) {
						throw (int) __LINE__;
					}
					
					// create list node with id and value
					list_node.first = m-1;
					list_node.second = strtod(str.c_str(),NULL);

					this->adjacency_list[n-1].push_back(list_node);

					// store highest global edge weight
					if (max_resistance < list_node.second) {
							max_resistance = list_node.second;
					}

					// store lowest global edge weight
					if (min_resistance > list_node.second) {
							min_resistance = list_node.second;
					}
					
				}

				// clear temp variables
				buffer.clear();
				c = '\0';
				n = 0;
				m = 0;
				str = ""; 
				
				// get next line
				getline(graph_string,line);	
				
			}

			if ( line.compare(0,3,"END") ) {
				throw (int) __LINE__;
			}

			graph_string.clear();
		}
		else {
			throw (int) __LINE__;
		}

	}
	catch (int e) {
		cerr << "There was an error loading graph, please check log file" << endl;
		throw e;
	}
}



// This function returns a vector where each element is a list of the vertices that
// belong to each groupN(i)  i= {0,..,max-degree}  
// Time: O(|V|)
vector< vector< int > > graph::getGroupN(void) const{

	// create group-N sets
	vector< vector< int > > groupN;

	try {

		// the groupN vector is filled with a list of integers. Where each integer represents
		// the vertices that belong to that groupN
		for (int i=0; i < (int) this->adjacency_list.size() ; i++ ) {

			// resize groupN in case the degree of the vertex is not yet represented in the groupN vector
			if (groupN.size() < this->adjacency_list[i].size()+1)
				groupN.resize( this->adjacency_list[i].size()+1);

			// the value of adjacency_list[i].size() represents the degree of each vertex
			groupN[this->adjacency_list[i].size()].push_back(i); 
		}



  	}
	catch (bad_exception&) {
		cerr << "There was an error on graph::getGroupN() line" << __LINE__ << endl;

	}

	return groupN;

}


// BFS algorithm to determine the distance(all edges treated as weight=1) between start and end.
// Time: O(|V|+|E|)
vector< int > graph::BFS(int start,int end) const{

	vector< int > geodesic;

	try {

		queue< int > next;
		vector< int > parent(this->vertices.size(),-1); //intialize the parent vector to -1
		vector< int > distance(this->vertices.size(),-1); //intialize the distance vector to -1
		bool found;
		int last_visited;

		found = 0;
		next.push(start);
		distance[start] = 0;
		last_visited = end;

		while (!next.empty() && !found) {

			// u is the current vertice
			int u = next.front();
			next.pop();
			
			for (int i=0; i < (int) this->adjacency_list[u].size()&& !found;i++) {
	     
				// Look through neighbors.
				int v = this->adjacency_list[u][i].first;
				
				// If v is unvisited.		
				if (distance[v] == -1) {
					distance[v] = distance[u] + 1;

					// mark the parent of v as vertex u
					parent[v] = u;

					if (v == end) found = 1;

					next.push(v);							
				}
			}
		}


		// travel the parent vector to build path
		while ( last_visited != -1 ) {
			geodesic.insert(geodesic.begin(),last_visited);
			last_visited = parent[last_visited];
		}
		
	}
	catch (bad_exception&) {
		cerr << "There was an error on graph::BFS() line" << __LINE__ << endl;
	}	
		
	return geodesic;

}

// Dijkstra algorithm to determine the shortest distance between start and end.
// Time: O(|E+V|*log|V|)
vector< int > graph::Dijkstra(int start,int end) const{

	vector< int > geodesic;

	try {

		priority_queue< pair< float,int >,vector< pair< float,int > >,greater< pair< float,int > > > Q; //min-priority queu
		vector< int > parent(this->vertices.size(),-1); //intialize the parent vector to -1
		vector< float > distance(this->vertices.size(),9.99e99); //intialize the distance vector to INF
		bool found;
		int last_visited;

		distance[start] = 0;
		Q.push(pair< float,int >(distance[start],start)); 
		found = 0;
		last_visited = end;

		while(!Q.empty() && !found ) {

            // fetch the best element of the queue (min distance)
            pair< float,int > top = Q.top();
            Q.pop();
                        
            // u is vertex index, d is the distance
            int u = top.second;
			float d = top.first;

			if ( u == end ) {
				found = 1;
			}

            if(d <= distance[u] && !found) {

				// iterate through all outcoming edges from u
				for (int i=0; i < (int) this->adjacency_list[u].size();i++) {

					int v = this->adjacency_list[u][i].first;
					float e = this->adjacency_list[u][i].second;
						
					// relax the vertex
					if(distance[v] > distance[u] + e) {
                    
						// update distance if possible
						distance[v] = distance[u] + e;
						// update the parent
						parent[v] = u;

						// add the vertex to queue
						Q.push(pair< float,int >(distance[v], v));

					}
				}
			}
		}


		// travel the parent vector to build path
		while ( last_visited != -1 ) {
			geodesic.insert(geodesic.begin(),last_visited);
			last_visited = parent[last_visited];
		}
		
		
	}
	catch (bad_exception&) {
		cerr << "There was an error on graph::Dijkstra() line" << __LINE__ << endl;
	}	
		
	return geodesic;

}


vector< vector< vector< vector< int > > > > graph::getGeodesics(void) const{

	vector< vector< vector< vector< int > > > > groupNGeodesics;

	try {
			vector< vector< int > > groupN;
			vector< vector< vector< int > > > groupNLongestGeodesics, groupNShortestGeodesics;
			vector< int > gtemp;
			vector< vector< int > > geodesics;
			int shortest, longest;

			groupN = getGroupN();
			groupNLongestGeodesics.resize(groupN.size()); 
			groupNShortestGeodesics.resize(groupN.size()); 

		
			// generate all possible combinations for every groupN set
			for (int i=0; i < (int) groupN.size(); i++) {

				shortest = longest = -1;

				for (int j=0; j < (int) groupN[i].size(); j++ ) {
					for (int k=j+1; k < (int) groupN[i].size(); k++ ) {
						

						// Run Dijkstra for each combination 
						// to get the geodesic between the two selected vertices
						gtemp = Dijkstra(groupN[i][j],groupN[i][k]);

						// check if the path is complete from start to end
						if (gtemp[0] == groupN[i][j] &&  gtemp[(int) gtemp.size() - 1] == groupN[i][k] ) {
							
							// add it to the geodesic list
							geodesics.push_back(gtemp);

							// store longest and shortest lengths
							if (shortest == -1)
									shortest = (int) gtemp.size();

							if (longest == -1)
									longest = (int) gtemp.size();

							if ((int) gtemp.size() < shortest ) 
								shortest = (int) gtemp.size();

							if ((int) gtemp.size() > longest ) 
								longest = (int) gtemp.size();
						}
			
					}
				}

				// add the shortests and longests geodesic to groupN i
				// go through all of the geodesics found and if length equal to longest or shorest add it
				for (int l=0; l < (int) geodesics.size(); l++ ) {

					if (geodesics[l].size() == shortest) 
						groupNShortestGeodesics[i].push_back(geodesics[l]); 
						
					if (geodesics[l].size() == longest) 
						groupNLongestGeodesics[i].push_back(geodesics[l]); 
						
				}

				//clear geodesics
				geodesics.clear(); 
			}


			//associate Longest and Shortest Geodesics to the groupNGeodesics array
			groupNGeodesics.push_back(groupNShortestGeodesics);
			groupNGeodesics.push_back(groupNLongestGeodesics);
		
	}
	catch (bad_exception&) {
		cerr << "There was an error on graph::getGeodesics() line" << __LINE__ << endl;
	}

	return groupNGeodesics;

}


vector< float > graph::findVoltages(int s,int t) const {
	
	vector< float > voltages;

	try {

		vector< int > system_variables, reverse_lookup(this->vertices.size(),-1);
		vector < float > conductance;
		vector< int > gtemp;
		float vertex_weight;
		int CurrentChildNode;

		TNT::Array2D< float > A;
		TNT::Array1D< float > x,b;
		
		voltages.resize(this->vertices.size());

		// initialize the voltages values
		for(int i=0; i < (int) voltages.size(); i++) {

			
			// if source, V(s) = 1.0
			if (i == s) {
				voltages[i] = 1.0;  
			}
			// if target, V(t) = 0.0
			else if (i == t) {
				voltages[i] = 0.0;  
			}
			// If the degree of the node is 1, then is treated as a ground (sink) node.
			else {
				
				// check if the node to be evaluated is connected from the source to the node using BFS
				// if not, not include at all. -- this will deal with graphs that are made of 
				// several non-connected subgraphs (in the case of the AIDS dataset)
				gtemp = BFS(s,i);

				// check if the path is complete from start to end
				if (gtemp[0] == s &&  gtemp[(int) gtemp.size() - 1] == i ) {
			
					if ( this->adjacency_list[i].size() == 1 ) {
						voltages[i] = 0.0; 
					}
					else {
						// this value is to identify the nodes that will form
						// the system of equations to solve using LU decomposition
						voltages[i] = 0.0; 

						// add the nodes that will form the system of equations
						reverse_lookup[i] = (int) system_variables.size();
						system_variables.push_back(i); 
						
					}
				}
				else {
					voltages[i] = 0.0;  
				}
			}
		}


		// initialize the system Ax= b.
		A = TNT::Array2D< float > ((int) system_variables.size(), (int) system_variables.size(),0.0);
		b = TNT::Array1D< float > (A.dim1(),0.0);
		x = TNT::Array1D< float > (A.dim1(),0.0);

		conductance.resize(system_variables.size()); 

		for (int i=0; i < (int) system_variables.size(); i++){
					
			conductance[i] = 0.0;

			for (int j=0; j < (int) this->adjacency_list[system_variables[i]].size(); j++){

				CurrentChildNode = this->adjacency_list[system_variables[i]][j].first;

				// read the edge value as a float 
				// inverse the edge weigth to use it as conductance, rather than resistance
				vertex_weight = 1.0 / this->adjacency_list[system_variables[i]][j].second;

				// fill the conductance(i) with the sum of all the edge weigths for vertex i
				conductance[i] += vertex_weight; 

				// initialize the b vector based on the nodes that have connections with the source
				if (CurrentChildNode == s) {
					b[i] = -1.0 * vertex_weight;
				}

				// exclude from adding to the matrix if is in the diagonal or if is ground (|adj(u)|=1, or is target)
				// or if source
				if (system_variables[i]!=CurrentChildNode && this->adjacency_list[CurrentChildNode].size() > 1 && CurrentChildNode != t && CurrentChildNode != s) {
					A[i][reverse_lookup[CurrentChildNode]] = vertex_weight;
				}
			}

			// Once the conductance of i has been calculated we can fill the diagonal
			A[i][i] = -1.0 * conductance[i];

		}


		//  solve the system of equations ONLY for the variables with unknown voltages
		//  Ax = b   Using LU decomposition
		
		JAMA::LU< float > lu(A);
		x = lu.solve(b); 

	
		// now that the voltages for the system variables have been calculated update
		// the original voltages values
		for (int i=0; i < (int) system_variables.size(); i++){
				voltages[system_variables[i]] = x[i];			
		}
			
	}
	catch (bad_exception&) {
		cerr << "There was an error on graph::findVoltages() line" << __LINE__ << endl;
	}

	return voltages;

}

graph::graphCurrents graph::getCurrents(map< string,double > &label_resistances) {

	graphCurrents groupNCurrents;

	try {

		/****************
		 groupNGeodesics looks like this:
			groupNGeodesics[0] == Shortest Geodesics
			groupNGeodesics[1] == Longest Geodesics
			
			Geodesics[0..groupN.lenght()], where each element is a vector of all of the geodesics that belong
			to Geodesics[i]  where i is the ith-groupN.
			Each geodesic is a vector of all of the vertices in it.
		*****************/

		vector< vector< vector< vector< int > > > > groupNGeodesics;
		vector< float > voltages;
		int s, t, geodesic_size;

		string nodePair, nodePairCurrent;
		stringstream temp_str;
		map< string,float > cachedResults;
		map< string,float > I;
		vector< float > Iout;
		float Itemp, vertex_weight;

		groupNCurrents.resize(2);
	
		groupNGeodesics = getGeodesics();

		groupNCurrents[0].resize(groupNGeodesics[0].size());  // shortest geodesic results
		groupNCurrents[1].resize(groupNGeodesics[1].size());  // longest geodesic results

	
		//modify the adjency list of each graph to add the resistance information of the labels
		for (int v=0; v < (int)  this->vertices.size(); v ++) {  
			for (int e=0; e < (int) this->adjacency_list[v].size(); e++  ) {

				this->adjacency_list[v][e].second += label_resistances[this->vertices[v].second];

			}
		}
			


		for (int g=0; g < 2; g++) {  //  process  Shortest then Longest Geodesics

			for (int i=0; i < (int) groupNGeodesics[g].size() ; i++) { // each ith-groupN

				groupNCurrents[g][i] = 0.0;

				for (int j=0; j < (int) groupNGeodesics[g][i].size(); j++ ) {  // each geodesic in ith-groupN
					
					//process each geodesic
					geodesic_size = (int) groupNGeodesics[g][i][j].size();

					s = groupNGeodesics[g][i][j][0]; // first vertex of the geodesic
					t = groupNGeodesics[g][i][j][geodesic_size-1]; // last vertex of the geodesic

					// create the key string for the cache map
					temp_str.str("");
					temp_str << s << "," << t;
					nodePair = temp_str.str();

					// check first if it's not cached
					if (!cachedResults[nodePair]) {  

						// based on the source and target, find the voltages of the circuit.
						voltages = findVoltages(s,t);

						// based on the voltages calculate the currents for each edge
						Iout.resize(this->vertices.size()); 

						for (int u=0; u < (int) this->vertices.size(); u++){

							Iout[u] = 0.0;

							for (int v=0; v < (int) this->adjacency_list[u].size(); v++) {

								// get the vertex weight
								vertex_weight = this->adjacency_list[u][v].second;

								// Current is equal to I = ( V(u) - V (v) ) / R(u,v)  
								Itemp = ( voltages[u] - voltages[this->adjacency_list[u][v].first] ) /  vertex_weight;

								// store only if current is downhill from u to adjacency_list[u][v].id
								if ( Itemp >= 0 ) {

									// create the current key string for the current map
									temp_str.str("");
									temp_str << u << "," << this->adjacency_list[u][v].first;
									nodePairCurrent = temp_str.str();

									I[nodePairCurrent] = Itemp;

									// store also in the Iout current vector
									// this is needed to calculate the prorated current thru the path
									Iout[u] += Itemp;
								}
							}
						}

						// now, with the currents for each edge, calculate the prorated current
						// thru path s,...,t

						Iout[s] = 1.0;  // The first current is direct from the source (not prorated)
						Itemp = 1.0;
			
						// pro-rate the current through the path. (If 0.0 then stop, since no current flows thru that path)
						for (int p=0; p < (int) groupNGeodesics[g][i][j].size() - 1 && Itemp != 0.0 ; p++) {

							// build the key to the node to access to read the current
							temp_str.str("");
							temp_str << groupNGeodesics[g][i][j][p] << "," << groupNGeodesics[g][i][j][p+1];
							nodePairCurrent = temp_str.str();

							// in case there is no current between two nodes, and the edge
							// is in the path, ignore the current between those two nodes
							if (  Iout[groupNGeodesics[g][i][j][p]] != 0.0 &&  I[nodePairCurrent]) {
								Itemp *= I[nodePairCurrent]/Iout[groupNGeodesics[g][i][j][p]];
							}

						}

						// cache the results to avoid calculating again
						cachedResults[nodePair] = Itemp;
					}


					// now the result should be available in cache, read from it, and store it
					// in the correspodent groupN for shortest and longest geodesics
					groupNCurrents[g][i] += cachedResults[nodePair];

					//clear any used maps
					I.clear(); 
					voltages.clear();
					Iout.clear();
					
				} // for-end    each geodesic within each ith-groupN

				// Now that all geodesics currents in the ith-groupN have been calculated create an average
				// in case there is more than one
				if ((float) groupNGeodesics[g][i].size() != 0 ) {
					groupNCurrents[g][i] /= (float) groupNGeodesics[g][i].size();
				}

			} // for-end     each ith-groupN
		} // for-end    shortest, longest 



	}
	catch (bad_exception&) {
		cerr << "There was an error on graph::getCurrents() line" << __LINE__ << endl;
	}
	
	return groupNCurrents;
}

graph::~graph(void) {
}
